var cluster = require('cluster');
var Cache = require('./cache-simple');

var errDesc = {
    '1': '[ERROR] sharedMemory.init(config) => config.manager is wrong type!'
};

/*\
    Manager
\*/

var Manager = function() {

    var self = this;
    
    // 初始化共享内存
    self.__cache__ = Cache.init();
    self.__sharedMemory__ = {};

    if(cluster.isMaster) {
        
        // manager是clusterMaster时，监听并处理来自worker的请求
        cluster.on('online', function(worker) {
            worker.on('message', function(data) {
                console.log('1', data);
                if (!data.isSharedMemoryMessage) return;
                self.handle(data);
                return false;
            });

            worker.on('message', function(data) {
                console.log('2', data);
            });
        });
    
    }else{
    
        // manager是clusterWorker时，监听并处理来自clusterMaster转发的请求
        process.on('message', function(data) {
            if (!data.isSharedMemoryMessage) return;
            self.handle(data);
        });

    }
};

Manager.prototype.handle = function(data) {
    var self = this;
    var value = this[data.method](data);

    var msg = {
        isSharedMemoryMessage: true,
        id: data.id,
        uuid: data.uuid,
        value: value
    };

    if(cluster.isMaster) {
        cluster.workers[data.id].send(msg);
        return;
    }

    process.send(msg);

};

Manager.prototype.setCache = function(data) {
    var __cache__ = this.__cache__;
    if(typeof data.value != 'undefined') {
        __cache__.set(data.key, data.value);
    } else {
        __cache__.remove(data.key);
    }
    return 'OK';
};

Manager.prototype.getCache = function(data) {
    return this.__cache__.get(data.key);
};

Manager.prototype.set = function(data) {
    this.__sharedMemory__[data.key] = data.value;
    return 'OK';
};

Manager.prototype.get = function(data) {
    return this.__sharedMemory__[data.key];
};


/* ========================================================
    User
======================================================== */

var User = function() {
    var self = this;
    self.__uuid__ = 0;
    self.__getCallbacks__ = {};
    process.on('message', function(data) {
        var cb = self.__getCallbacks__[data.uuid];
        if (cb && typeof cb == 'function') {
            cb(data.value)
        }
        self.__getCallbacks__[data.uuid] = undefined;
    });
};

User.prototype.uuid = function() {
    var i = this.__uuid__;
    return this.__uuid__ = i = i > 65535 ? 0 : i + 1;
};

User.prototype.set = function(key, value, callback) {
    this.handle('set', key, value, callback);
};

User.prototype.remove = function(key, callback) {
    this.set(key, undefined, callback);
};

// eg. callback = function(data) {};
User.prototype.get = function(key, callback) {
    this.handle('get', key, null, callback);
};

User.prototype.setCache = function(key, value, callback) {
    this.handle('setCache', key, value, callback);
};

User.prototype.removeCache = function(key, callback) {
    this.setCache(key, undefined, callback);
};

User.prototype.getCache = function(key, callback) {
    this.handle('getCache', key, null, callback);
};

User.prototype.handle = function(method, key, value, callback) {

    var self = this;
    var uuid = self.uuid();

    process.send({
        isSharedMemoryMessage: true,
        method: method,
        id: cluster.worker.id,
        uuid: uuid,
        key: key,
        value: value
    });

    self.__getCallbacks__[uuid] = callback;

};


/* ========================================================
    Transfer
======================================================== */

var Transfer = function(whoIsManager) {
    cluster.on('online', function(worker) {
        worker.on('message', function(data) {
            if (!data.isSharedMemoryMessage) return;
            if (worker.id === whoIsManager) {
                // 转发给user
                cluster.workers[data.id].send(data);
                return;
            }
            // 转发给manager
            cluster.workers[whoIsManager].send(data);
        });
    });
};

/* ========================================================
    Init
======================================================== */

function init(config) {

    config = config || {};

    var defaultConfig = {
        manager: undefined,      // undefined => master | cluster_id => cluster
        max: 10 * 1024 * 1024,   // 单位: byte
        expire: 30 * 60 * 1000
    };

    var whoIsManager = config.manager || defaultConfig.manager;

    /*
        Manager(cluster worker) <--> Transfer(cluster master) <--> User、User、User...(cluster worker)
    */
    if (whoIsManager) {

        if(typeof whoIsManager !== 'number') {
            throw new Error(errDesc[1]);
            return;
        }
        
        if (cluster.isMaster) {
            return new Transfer(whoIsManager);
        }
        
        if(cluster.worker.id === whoIsManager) {
            return new Manager;
        }
        
        return new User;

    } else {

        /*
            Manager(cluster master) <--> User(cluster worker)
        */        

        if (cluster.isMaster) {
            return new Manager();
        }

        return new User;

    }

}

exports.init = init;

